by
Parallel Contraction Hierarchies Can Be
Efficient and Scalable
Abstract.
Contraction Hierarchies (CH) (Geisberger et al., 2008) is one of the most widely used algorithms for shortest-path queries on road networks. Compared to Dijkstra’s algorithm, CH enables orders of magnitude faster query performance through a preprocessing phase, which iteratively categorizes vertices into hierarchies and adds shortcuts. However, constructing a CH is an expensive task. Existing solutions, including parallel ones, may suffer from long construction time. Especially, in our experiments, we observe that existing parallel solutions demonstrate unsatisfactory scalability, and have performance close to sequential algorithms.
We present SPoCH (Scalable Parallelization of Contraction Hierarchies), an efficient and scalable CH construction algorithm in parallel. To address the challenges in previous work, our improvements focus on both redesigning the algorithm and leveraging parallel data structures. We compare SPoCH with the state-of-the-art sequential and parallel implementations on 16 graphs of various types. Our experiments show that SPoCH achieves – speedups over the best sequential baseline and – speedups over the best parallel baseline in CH construction, while maintaining competitive query performance and CH graph size. We released our code and all datasets used in this paper.
1. Introduction
Computing shortest distance is one of the most fundamental graph problems, playing a vital role in various applications, such as navigation on road networks. To accelerate point-to-point distance queries, many existing solutions use two-phase approaches, which preprocess the graph and construct an index—often an auxiliary graph—to facilitate queries. Among the most notable two-phase solutions is the Contraction Hierarchies (CH) (Geisberger et al., 2008), which is mainly designed for sparse networks such as road networks, and is widely used in practice, for instance in Google Maps. In addition to being used on its own, CH is also a vital component in other approaches for various applications on distance queries, such as Transit Node Routing (Arz et al., 2013; Bast et al., 2006, 2007), Hub-Based Labeling (Abraham et al., 2011, 2012; Delling et al., 2013a), and some renowned algorithms including CHASE (Bauer et al., 2010) and PHAST (Delling et al., 2013b). We refer the audience to the excellent surveys (Sommer, 2014; Bast et al., 2016; Madkour et al., 2017) for more background of the state-of-the-art techniques for route planning.
| Algorithms | Preprocessing time | Query time |
| Dijkstra (Dijkstra, 1959) | - | 7.20 s |
| -stepping∗ (Dong et al., 2024b) | - | 0.320 s |
| RoutingKit (contributors, 2024; Dibbelt et al., 2016) | 2466 s | 79.1 s |
| PHAST (Group, 2024; Delling et al., 2013b) | 1341 s | 138 s |
| CH-Constructor∗ (Bühler and Nusser, 2024) | 1527 s | 317 s |
| OSRM∗ (Luxen and Vetter, 2024, 2011) | s | 163 s |
| Ours∗ | 23.1 s | 93.3 s |
In this paper, we consider a graph with an edge weight function . Given two vertices , a distance query asks for the shortest distance from to . The idea of CH, as illustrated in Fig. 1(a), is to “contract” the graph into a hierarchy, which is an auxiliary graph that preserves the shortest distances in , such that distance queries on can be much faster. In each iteration, one vertex and all its edges are moved from the original graph to an auxiliary graph , forming a level in . Additionally, extra edges (shortcuts) are added to the original graph, forming the overlay graph , to preserve the pairwise distances among the remaining vertices. Finally, all vertices in form the contraction hierarchies (CH). A distance query requires a bidirectional search on .
Empirically, a distance query on the CH touches far fewer vertices than in the original graph, thereby resulting in much faster query speed. As shown in Tab. 1, for the North American roadmap from OpenStreetMap (OpenStreetMap contributors, 2010), queries can be – faster than directly using SSSP algorithms, such as Dijkstra (Dijkstra, 1959) and -stepping (a parallel SSSP algorithm) (Meyer and Sanders, 2003; Dong et al., 2021). However, this impressive querying speed comes at a cost—constructing is expensive. Existing sequential solutions use 22–41 minutes, and the best parallel solutions need more than 300 seconds. Such long preprocessing time may require significant computational resources, and limit its adaptability to large graphs. Such running time is reasonable for a sequential algorithm—both RoutingKit (Dibbelt et al., 2016) and PHAST (Delling et al., 2013b) take construction time than Dijkstra, which is acceptable considering the speedup for queries. However, the existing parallel CH algorithms are only faster than (sequential) PHAST even on a 96-core machine. The results indicate a significant gap, suggesting the potential for improvements in parallelism in CH construction algorithms.
In this paper, we propose SPoCH (Scalable Parallelization of Contraction Hierarchies), which supports scalable and efficient contraction hierarchies construction with high parallelism without compromising the query performance. The performance gain is mainly from good parallelism and algorithmic improvements for sequential performance.
Conceptually, parallelizing CH construction is not hard. In 2009, Vetter (Vetter, 2009) pointed out that vertices in an independent set (i.e., vertices that do not share edges) can be contracted in parallel, as illustrated in Fig. 1(b). However, despite being studied in many later papers (Luxen and Vetter, 2024, 2011; Karimi et al., 2019, 2020; Kieritz et al., 2010; Bühler and Nusser, 2024), we are unaware of scalable multicore CH implementations. In fact, constructing CH is a sophisticated process with intensive computation. Achieving good performance requires high parallelism in all steps and careful algorithmic design. We observe two major challenges. First, identifying the vertices to contract involves simulating contractions on many (if not most) vertices, which calculates distances for numerous vertex pairs. Such a process is expensive, requiring high parallelism to enable good performance. The second challenge is to dynamically maintain the graph in parallel, since the graph may experience rapid changes, such as the removal of contracted vertices/edges and the insertion of shortcuts. Existing solutions may sacrifice parallelism (e.g., using locks) to support such dynamic updates. Ideally, a scalable solution would necessitate graph maintenance support with full parallelism.
Contributions. We propose SPoCH, which consists of a new algorithmic framework for CH, and a high-performance implementation. Our improvement lies in both algorithm and data structure design with high parallelism. Algorithmically, we introduce the new LocalSearch step in construction to address the high cost of simulated contraction on a large fraction of vertices. This step gathers all the pairs of vertices for distance computation, and processes them in a batch in parallel. Processing them as a batch allows for combining and eliminating repeated computations, offering high potential to exploit parallelism, as well as a pruning process to remove suboptimal shortcuts added in previous rounds. Finally, SPoCH memoizes the computed distances in this step to facilitate later computation. We provide more information in Sec. 3.1. Combining these benefits, the LocalSearch step improves both the sequential running time and parallelism.
Another key performance improvement in SPoCH comes from using efficient parallel data structures. As mentioned, to get a highly parallel solution, we need to efficiently handle edge updates to the graph in parallel. We use the phase-concurrent hash table (Shun and Blelloch, 2014) to buffer newly added shortcuts in each round, which allows for efficient lock-free concurrent insertions of new edges. However, combining these edges to the overlay graph in each round requires scanning the entire hash table array and the edge list of the overlay graph. This may be inefficient in most rounds where only a small number of shortcuts are added. Our solution employs a lazy update scheme to delay the combination of the shortcuts to the overlay graph, while still making the delayed shortcuts visible to future computations. We introduce this technique in Sec. 4. This approach reduces the cost of maintaining the graph from to of the overall running time on average across all tested graphs.
We compare SPoCH with the state-of-the-art sequential and parallel solutions on various graphs. Even the sequential running time of SPoCH is as fast as the sequential baselines. Due to our new design, SPoCH achieves high parallelism and good scalability with large numbers of processors (see Fig. 6). On North America in Tab. 1, SPoCH is – faster in construction than the baselines while achieving similar query performance. On the 16 graphs we tested in Tab. 4, compared to the fastest baseline on each graph, SPoCH is 3.83–41.0 faster, with an average of . We also conduct in-depth performance studies of the CH algorithms in Sec. 6. We released our code in (Wan et al., 2024).
Preliminaries. We focus on the shared-memory setting with fork-join parallelism. The computation starts with one thread. A thread can fork two child software threads to work in parallel. When both children complete, the parent thread continues. Such a computation can be executed by a randomized work-stealing scheduler (Blumofe and Leiserson, 1999; Arora et al., 2001; Blelloch et al., 2020b). We use the atomic operation compare_and_swap, which checks if the memory location pointed to by has the value , and if so, changes the value to . The function returns true if it successfully changes the value, and false otherwise. We also use write_min, which reads the memory location pointed to by and writes value to it if is smaller than the current value.
We consider a weighted graph with , , and an edge weight function . For , we define as the in-neighbors of , and as the out-neighbors of . We use . For a set of vertices , we define as all neighbors of vertices in . Note that the main process of the CH construction algorithm is to update the overlay graph by removing contracted vertices and incident edges, and inserting shortcuts. Therefore, when we use in the algorithm description, we always refer to the neighborhood of a vertex on the overlay graph.
| : The input graph | |
| : The overlay graph | |
| : Weight for edge | |
| : The score of vertex | |
| : In-neighbors of vertex | |
| : Out-neighbors of vertex | |
| : In- and out-neighbors of vertex | |
| (Unless otherwise specified, we use , and to refer | |
| to the neighbors of on the overlay graph) | |
2. Contraction Hierarchies
Contraction Hierarchies (CH) was proposed by Geisberger et al. (Geisberger et al., 2008) in 2008, building upon the concept of highway hierarchies (Sanders and Schultes, 2006; Knopp et al., 2007). CH is one of the most widely adopted techniques in route planning and is used in many systems, such as Google Maps.
Let be the input graph. Its Contraction Hierarchies is an auxiliary graph that preserves the pairwise distances of . Distance queries on can be orders of magnitude faster than on itself, particularly for sparse graphs such as road networks. The CH algorithm iteratively contracts the “least important” vertex (determined by a user-specified score function) from the graph, and adds its edges to the contraction hierarchy. Additional edges (shortcuts) are added to the remaining graph (referred to as the overlay graph ) to preserve the distances. Each contracted vertex represents a level in the CH, forming a hierarchy. This process is illustrated in Fig. 1(a). A distance query from to is performed by running a bidirectional search from both and on the CH. The query algorithm is presented in Sec. 5. By contracting vertices and adding shortcuts, the hop distances between vertices are greatly reduced, allowing for much faster query performance.
Given its wide range of applications (Funke et al., 2019, 2015; Abeywickrama et al., 2016; Sanchez et al., 2022; Ouyang et al., 2020), CH has been extensively studied. In this section, we first review the algorithm proposed by Geisberger et al. (Geisberger et al., 2008) along with the fundamental concepts. Then in Sec. 2.2, we introduce existing parallel solutions. Other related studies are discussed in Sec. 7.
2.1. Sequential Solutions
We present the sequential construction algorithm for CH in Alg. 1. The algorithm takes a graph as the input, and computes the CH for in . As mentioned, the contraction process requires removing (contracting) vertices, and adding shortcuts to preserve the shortest distances. To avoid destroying the input graph, an overlay graph is used to reflect the changes. A score array is also maintained for all uncontracted vertices, representing their “importance” and deciding the contracting order.
The algorithm starts by computing the initial scores of all vertices (line 1). Then, it iteratively selects and contracts the vertex with the lowest score (line 1) until the overlay graph is empty (line 1). When contracting a vertex , the algorithm computes the shortest paths (line 1) from each to each by running Dijkstra’s algorithm. These shortest path queries are referred to as the Witness Path Search (WPS) (Geisberger et al., 2008; Karimi et al., 2019). If the shortest path length from to is the same as (line 1), then can be on the shortest path from to . Thus, the shortcut edge with weight is added to to preserve the distances on (line 1 and 1). After that, and all its incident edges are removed from the overlay graph , and are added to the CH . Contracting may cause the score of all its neighbors to change, which are updated on line 1. Finally, the algorithm moves on to contract the next vertex, until all vertices are contracted.
Vertex Score and Edge Difference. Scoring the vertices is the most time-consuming part of CH construction and significantly impacts the shortest path query performance in CH. The most widely used score function is the edge difference (Geisberger et al., 2008). For a vertex , the edge difference is the change in the number of edges on the overlay graph after contracting , i.e., the number of shortcuts added minus the number of edges incident to . A vertex with a higher edge difference is considered more important because its contraction has a larger impact on the graph’s structure. To compute the number of shortcuts to be added, a simulated contraction on is required, which virtually generates and counts the shortcuts. This simulation is very similar to line 1–1 except it does not add edges or modify edge weights in .
In existing work (Geisberger et al., 2008, 2012), the score function is usually defined as a weighted combination of edge difference and many other metrics, such as vertex degree, hop distance, and hierarchy depth. Among them, edge difference has the highest weight and is usually the most computationally expensive part. Therefore, for simplicity, our algorithm description uses edge difference as the score function. Our algorithmic ideas are independent of the score function.
2.2. Existing Parallel Algorithms
The construction for CH is very costly, so it is natural to consider parallelizing this process by contracting many vertices in parallel. However, note that we cannot contract arbitrary vertices in parallel. For instance, if two adjacent vertices and are contracted together, may choose to add a shortcut incident to . Since is removed from the overlay graph at the same time, the shortcut may not be added correctly, resulting in distances not being preserved accurately on the overlay graph. In 2009, Vetter (Vetter, 2009) first observed that vertices that do not share edges (i.e., an independent set) can be contracted in parallel, and proposed the first parallel CH construction algorithm. In this case, all added shortcuts will be connected to uncontracted vertices and are thus safe to add to the overlay graph.
Almost all later parallel solutions (e.g., (Karimi et al., 2020; Bühler and Nusser, 2024; Luxen and Vetter, 2011; Kieritz et al., 2010; Karimi et al., 2019)) follow Vetter’s high-level idea. In each round, an independent set (IS) of vertices is identified to be contracted. All these vertices will be placed on the same level in the CH. After that, shortcuts will be added normally. This process is repeated until all vertices have been contracted. As with the sequential algorithm that first contracts vertices with the lowest score, the vertices in the IS should also have low scores in general. In Vetter’s original algorithm, the IS includes all vertices with the smallest score within its -hop neighborhood. OSRM (Luxen and Vetter, 2024, 2011) specifically uses . Our implementation also follows this idea and uses . An existing GPU algorithm (Karimi et al., 2019) and CH-Constructor (Bühler and Nusser, 2024) find a maximal independent set (MIS) and include all vertices in this MIS with scores smaller than a threshold.
Challenges to Achieve High-Performance Implementations. While Vetter’s work reveals the potential parallelism in the algorithm, the idea itself is not sufficient to enable a high-performance solution. Many challenges remain in other parts of the algorithm. We highlight two components here. The first part is the process to (re)compute the scores of vertices. When multiple vertices are contracted together, a large number of simulated contractions is needed, which all involve running WPSes by Dijkstra’s algorithm. The second challenge is to update the overlay graph in parallel, since contracting vertices in parallel results in a bulk of new shortcut edges that need to be added to the overlay graph. Both parts (performing WPS and updating the graph) involve expensive computation, and require careful algorithm design and parallelization. In our experiments, we test two existing parallel solutions CH-Constructor (Bühler and Nusser, 2024) and OSRM (Luxen and Vetter, 2024, 2011) based on Vetter’s algorithm. In tests on a 96-core machine with 16 graphs (Tab. 4), their performance is only up to faster than a highly-optimized sequential implementation, and can be up to 3.59 slower on some graphs.
In the next section, we present our solution that achieves a scalable and efficient parallel CH construction algorithm.
3. Our Parallel CH Construction Algorithm
In this section, we present SPoCH for generating contraction hierarchies efficiently in parallel. As with other parallel CH algorithms, SPoCH also follows the high-level idea from Vetter (Vetter, 2009) that contracts a batch of vertices in an IS in parallel. To achieve high parallelism without sacrificing CH quality, SPoCH uses several novel techniques that will be discussed in this section.
We present the pseudocode of SPoCH in Alg. 2. It takes a graph as the input, and returns the contracted graph . Note that the vertex set in is the same as in the input graph, and the algorithm only needs to compute the edges of the contracted graph. Similar to Alg. 1, we also maintain the overlay graph to reflect the changes in the original graph due to contractions.
The main loop of our algorithm (the while-loop on line 2) repeatedly finds an IS of vertices and contracts them. We summarize this process in five steps. We will briefly outline the high-level idea here, and then elaborate on each step.
One of the key insights in SPoCH is the introduction of the LocalSearch step on line 2. This step is used to reduce redundant WPSes, create more parallel tasks, and overall reduce computation while improving parallelism. With the details of this step described in Sec. 3.1, the overall goal of this step is to preprocess the current overlay graph by running WPSes from a set of vertices , in order to (re)compute the score for a set of vertices . At the beginning, both and are . In later rounds, will be computed by the Contract step in the previous round, and are the neighbors of vertices contracted in the previous round. In Vetter’s algorithm, WPSes are directly performed when recomputing scores in each round. SPoCH separates this part with new designs, and further utilizes the results for more optimizations.
The next two steps Score and Select will (re-)calculate the score for vertices in , and choose an independent set of vertices to contract. To ensure that selected vertices have low scores, includes all vertices with the lowest score in their neighborhood. Importantly, the Score step makes use of the results from the LocalSearch step to avoid redundant computation.
Finally, the Contract step performs the contraction and updates the overlay graph, which is another major improvement over existing solutions. In this step, all incident edges to are moved to , and the corresponding shortcuts are added to . All these updates in SPoCH are highly parallel. Especially, we use lock-free data structures to update the graph efficiently. We introduce the algorithmic idea in Sec. 3.4, and discuss the parallel data structure support in Sec. 4. In addition, to facilitate the LocalSearch step, the Contract step will generate the set of vertices , which are all vertices from which we will start WPS in the next round.
| : | Vertices to be contracted | |
|---|---|---|
| : | Vertices that need to | in the previous round |
| recalculate their scores | ||
| : | Shortcuts to be added | |
| : | Starting points of shortcuts | |
| : | Set of WPS sources |
We summarize the notations in our algorithm in Tab. 3. Next, we will elaborate on each step in Alg. 2. Later, in Sec. 4, we discuss using efficient data structures to support this algorithm.
3.1. Step 1: Local Search with Memoization
One improvement of SPoCH lies in the stand-alone LocalSearch step with memoization at the beginning of each round, presented in Alg. 3. This step runs witness path searches (WPSes) to prepare the shortest distances needed during scoring (Sec. 3.2) and contracting (Sec. 3.4). As mentioned in Sec. 2.1, to compute the score of a vertex , a simulated contraction on is needed to determine the shortcuts that would be added if we contract . For each pair of vertices where and , if is on the shortest paths from to , then a shortcut is needed. We also use the WPS as described in Sec. 2.1, which runs Dijkstra from until is settled to find the shortest distance between them.
Running WPSes is crucial to determine the score of a vertex. However, it is also expensive—to compute the edge difference for a vertex , WPSes are needed on all pairs in (Cartesian product). Fully finishing WPSes for all pairs may be expensive, especially when there exists one pair of vertices that are extremely far away. Therefore, most existing systems (including ours) also limit the size of each WPS to be , i.e., each Dijkstra’s algorithm only searches for iterations. In this case, some suboptimal shortcuts may be added and make the CH (necessarily) larger. To improve the overall performance, we highlight three key techniques in our LocalSearch step: batching, memoization, and pruning.
Batching. Our first technique is to collect all WPSes needed and process them in a batch. In SPoCH, the Contract step generates a vertex set , which serves as sources of WPSes in the next round. By maintaining them in a batch, we process all WPSes in a batch to avoid redundant computation and achieve better parallelism. For example, if two simulated contractions trigger WPSes from the same vertex, they will be performed only once in the batch. An illustration of all WPS sources and how batching saves the number of WPSes is given in Fig. 2. Executing all WPSes as one batch allows us to optimize them as a whole, and provides greater opportunity to exploit parallelism.
Memoization. SPoCH uses memoization to avoid computing the shortest distance on the same vertex pair multiple times. In SPoCH, a map is used to memoize pairwise shortest distances computed in previous WPSes. If a WPS is required on the same pair again, we directly use the memoized results without running another WPS. This information is reused in pruning (described below) and later steps Score and Contract.
Pruning. A byproduct of memoizing the distances in is to prune redundant edges in . As the algorithm proceeds, more pairs of vertices obtain their true shortest distance by WPSes (using Dijkstra’s algorithm). If there is an edge between and in the current overlay graph (possibly a suboptimal shortcut added in previous rounds) with weight , the edge can be deleted since it is redundant. This optimization effectively reduces the size of the overlay graph, which potentially improves the performance for both construction and query.
The LocalSearch Step. With the three techniques, we now describe the LocalSearch step. The step takes , and as input, and performs WPSes from all vertices in to recompute the scores for all vertices in . For each WPS from , we first identify the set of vertices that need to be settled by . On line 3–3 in Alg. 3, for each , if , this means that is a source of WPS because is an in-neighbor of an affected vertex . To compute the new score for , the shortest distances from to each vertex need to be computed. Therefore, we collect , the out-neighbors of , to the set . Note that can correspond to multiple affected vertices . In this case, all of their out-neighbors will be gathered in , and only incur one search starting from . Before adding a vertex to , we also check if has already been memoized in . If so, will be skipped.
With all target vertices of WPS originating from gathered in , We run Dijkstra’s algorithm from to search for all vertices in . During the algorithm, the shortest distances from to each are inserted into . As mentioned, a WPS terminates either when all vertices in have been settled, or when vertices have been settled (so that the overall cost is limited). If the WPS terminates before a vertex is settled, the distance between and will be viewed as and a shortcut between and is always generated.
3.2. Step 2: Score
The second step computes the scores for each vertex , as presented in Alg. 4. In the previous step, we have memoized the distances in the neighborhood of each in a map . To recompute the score of , we first set to the negative of its number of direct neighbors, i.e., , since edges can be removed from if gets contracted (line 4). Then, we iterate through each pair of in-neighbors and out-neighbors of and check if their distance through is equal to or smaller than the shortest distance stored in . If so, we add one to since one shortcut needs to be added.
For simplicity, our pseudocode only computes the edge difference, which is the most important component of the score for each vertex. In practice, many other components are considered in existing work. SPoCH uses a similar score function as in previous work, which is a combination of edge difference, vertex degree, hop distance, and hierarchy depth.
3.3. Step 3: Select
The Select step identifies a subset of vertices to contract in parallel, as presented in Alg. 5. We aim to contract multiple vertices in an independent set (i.e., do not share edges) simultaneously while prioritizing those with lower scores. To do this, we select all vertices that have the minimum score in its neighborhood. As shown in line 5–5 in Alg. 5, we process each vertex in . If has the lowest score among its neighbors, we insert into a set . To handle equal scores, we use the label of each vertex to break ties. This approach can find an independent set so that vertices with lower scores are contracted before those with higher scores. Fig. 1(b) presents an example of contracting a graph in parallel. In each round, multiple independent vertices can be contracted together, and the levels can be determined accordingly. The final graph preserves the shortest distance between any two vertices.
Selecting an IS of vertices ensures that shortcuts are always established between non-contracted vertices. The process is also reasonably fast for contraction. In our experiments, we observe that usually within 10–20 rounds, more than 99% of vertices are contracted, indicating high potential of parallelism in this algorithm.
3.4. Step 4: Contract
With the independent set computed in the previous round, the last step Contract performs the actual contraction, as given in Alg. 6. This step moves all vertices in from to the CH, adds incident edges to , computes and adds the relevant shortcuts to , and finally generates the set as the sources for running WPSes in the next round, which will be used in the next LocalSearch step.
In Alg. 6, we process each vertex in parallel. We first remove each vertex from the overlay vertex set in line 6. Next, as shown in line 6–6 in Alg. 6, we move the incident edges of to , which contains two subsets: and . When moving a vertex to the CH, its incoming edges are stored in , and outgoing edges are stored in . This separation is necessary because during the shortest path queries, the search from the target only proceeds “backward” by following incoming edges in , while the search from the source only moves “forward” by following outgoing edges in . Both searches move “upward” in the CH.
Finally, as shown in line 6–6 in Alg. 6, for each vertex to be contracted, we enumerate each in-neighbor and out-neighbor . We establish a shortcut between and if necessary. This is performed by comparing with the shortest distance stored in . Recall that is a map generated in the LocalSearch step, which buffers shortest distances for relevant vertex pairs. If , then the shortest path between and can be via . A special case is when is not in , which means that the shortest distance between them was not computed in the LocalSearch step. In both cases, before contracting , a shortcut needs to be established between and with weight (line 6).
However, a write–write race condition may occur when multiple shortcuts that share the same source or target node are inserted in parallel. An example of this issue is shown in Fig. 3. Vertex has three out-neighbors that are being contracted, each introduces a new out-neighbor to . Since the contractions of these three original out-neighbors, , run concurrently, the insertions of ’s new out-neighbors, , also happen in parallel. Therefore, we need a parallel data structure that supports frequent and concurrent insertions, while providing fast query performance. In our implementation, we use the parallel hash map in Sec. 4.1 to buffer all shortcuts added in this round and later combine it with on line 6. In Sec. 4.2 we will describe other optimizations to reduce the cost to combine and . Finally, we can also remove all edges incident on from after adding shortcuts. This is because these edges no longer contribute to the reduction of distances to any vertices that remain uncontracted.
During the contraction step, we also generate the set that contains all sources for the WPSes in the next step. When contracting a vertex , we check all and to see whether a shortcut from to needs to be established. If so, the scores for both and are affected, since both have their neighborhoods changed. To reconsider the score of , WPSes are needed from (to find the distance from the new in-neighbor to all ). To reconsider the score of , WPSes are needed from all in-neighbors of (to find their distances to the new out-neighbor ). Hence, we first use a set to buffer all vertices . At the end of this step, we construct the set , which includes all vertices in and all in-neighbors of them. This set will be used as the sources for the WPSes in the LocalSearch step in the next round. An illustration of the vertex sets and is shown in Fig. 2.
3.5. Postprocessing
After contracting all vertices, we re-index and sort the vertices (Blelloch et al., 2010; Dong et al., 2024a) by their levels and scores to improve query performance. In the sequential CH algorithm (Geisberger et al., 2008), vertices are reordered by the order in which they are contracted. In PHAST (Group, 2024), vertices are grouped by levels to perform SSSP queries on the CH, where lower level vertices have a lower rank than those in higher levels. We combine the two re-indexing methods and use vertex scores to approximate the contraction order. For any two vertices, if they are at the same level, they are sorted by score from the lowest to the highest. Otherwise, they are sorted by level from the lowest to the highest.
This re-indexing ensures that the vertices are efficiently organized for query processing, where paths are traversed based on the contraction rank. Once the vertices are re-indexed, we generate the final CH from and .
4. Parallel Data Structures
In Sec. 3, we presented the algorithmic idea of SPoCH. We note that, however, to support high performance in practice, the implementation relies on efficient data structures to support the algorithm, especially in handling dynamic updates of the overlay graph. Note that the (overlay) graph is highly dynamic—as the contraction proceeds, the graph consistently undergoes vertex removals (removing contracted vertices), edge removals (removing edges incident on contracted vertices), and edge insertions (adding shortcuts). Each of these operations require high parallelism to enable good performance for the overall CH algorithm. This section provides an overview of the parallel data structures used in our implementation.
Especially, to facilitate graph traversal, the overlay graph is usually maintained by an array-based structure (e.g., the CSR introduced below), which maintains the neighbors of the same vertex in a consecutive chunk in the memory. When multiple vertices are contracted in parallel, shortcuts are also generated by multiple threads. Inserting them into the overlay graph in a concurrent manner is therefore challenging. In existing implementation, OSRM avoids such complications by only contracting vertices with the minimum score in its 2-hop neighborhood. In this case, for a vertex in the overlay graph, only one contracted vertex may update the neighbor list of , and this will be performed sequentially. However, by restricting all contracted vertices to be at least two hops away, OSRM can choose far fewer vertices in the independent set in each round. In our experiments, we observe that such restriction greatly increase the number of rounds and degraded the performance. In CH-Constructor, a lock-based data structure is used, which also limits parallelism. In this section, we introduce the parallel data structures used in SPoCH, such that they support updating the overlay graph in an efficient and flexible way.
4.1. Background of Existing Data Structures
SPoCH mainly uses two data structures to maintain graphs: Compressed Sparse Row (CSR) (Tinney and Walker, 1967) and phase-concurrent hash table (Shun and Blelloch, 2014). In this section, we introduce the background of these data structures and their interface. Then in Sec. 4.2, we describe how they are used to support our algorithm.
Compressed Sparse Row (CSR) (Tinney and Walker, 1967). CSR is a widely used array-based adjacency list representation in graph algorithms (Dong et al., 2021, 2023; Liu et al., 2025; Wang et al., 2023; Dhulipala et al., 2024; Dong et al., 2024b; Dhulipala et al., 2021; Wheatman et al., 2024), consisting of an edge array of size and an offset array of size . The edge array is a concatenation of neighbor lists (along with the corresponding edge weights) for all vertices. The neighbor list for each vertex is always contiguous. The offset array stores the start position of each vertex’s neighborhood in the edge array. CSR provides fast lookups with good cache locality due to its contiguous memory layout. However, it does not support insertion or deletion without rebuilding the entire edge and offset arrays. Our input graph is given in the CSR format.
Phase-Concurrent Hash Table (Shun and Blelloch, 2014). The phase-concurrent hash table is an array-based structure that supports efficient operations for an unordered set, including insertions, deletions, and lookups. We use it to support multiple variables in SPoCH.
Unlike traditional concurrent hash tables that allow arbitrary operations of any type to run concurrently, the phase-concurrent hash table only permits operations of the same type to proceed simultaneously. This design ensures both serialization and determinism (w.r.t. execution order). To insert a key-value pair into the hash table , the key is first hashed to a random index in . If this index is occupied, linear probing is performed to find the next available slot. Once an empty index is found, it swaps in the new record using an atomic operation compare_and_swap. When multiple threads attempt to modify the same memory location, the atomic operation guarantees that only one thread succeeds. All other threads will fail in the operation and continue linear probing. A special case for insertion is to update the value for the if it already exists in the hash table. In our implementation where the value can refer to the weight of a given edge, we update the value if the newly inserted value is smaller than the current value stored in the hash table. In particular, when inserting , if we find that is already in the table during linear probing, we use an atomic operation write_min to update the value and keep the lower one. The atomicity of this operation guarantees that only the minimum value among all concurrent modifications is retained.
Similarly, to look up a record with key , the hash table hashes to an initial index . Then, three cases are considered: 1) If is empty, then does not exist in the hash table. 2) If the key at is , then the record with key has been found. 3) If the key at is not , the algorithm uses linear probing to continue searching for the key in subsequent indices, until is found or an empty slot is encountered. To delete a record with key , the algorithm first locates in . If is present, the corresponding index is marked as a tombstone, which is not treated as empty but will not match any existing keys. If is not found in , no further action is needed.
For simplicity, we will refer to the phase-concurrent hash table as the hash table in the following part of this section.
4.2. Using Parallel Data Structures in SPoCH
Maintaining the Overlay Graph Dynamically. To facilitate graph traversal, the overlay graph is maintained using a CSR, such that getting the neighbor list for each vertex is always efficient. The challenge is then to efficiently reflect edge insertions in the static CSR structure.
To efficiently support these updates in the overlay graph, we use an auxiliary edge set to tentatively maintain the newly added edges. Recall that when a node is contracted, pairwise edges from to need to be added to the overlay graph if the shortest path from to passes through . Instead of adding them directly to the CSR of the overlay graph, we first use to buffer them. Once a shortcut is generated (line 6 in Alg. 6), we directly add the new edge to . However, adding these edges to in parallel is challenging due to contention—although contracting two adjacent vertices is avoided by selecting an IS, it is possible for multiple vertices to be contracted share the same neighbor. In this case, new edges will be concurrently added to this neighbor. Moreover, edges with the same endpoints could be added to the overlay graph by two different contracted vertices and (i.e., via –– and ––). Among all the same edges, only the one with the smallest weight should be kept.
To support these updates in the overlay graph, we use a parallel hash table to maintain . Once a shortcut with weight is generated, we insert the triple to the hash table. To insert an edge , we first check whether the edge exists in the hash table. If so, we use write_min to update the weight atomically if is smaller than the stored weight. Otherwise, we insert to the hash table. Here we use as the key, and we will explain the design choice later in this section.
Note that even if we first collect shortcuts in and combine them with to avoid frequently updating the CSR, it can still be expensive to do so every round. Especially in later rounds, only a few vertices are contracted, generating a small number of shortcuts. In this case, updating the CSR to add a few edges is inefficient. To handle this issue, we combine the two sets in a lazy manner. In particular, we do not clear and combine with until gets overloaded (i.e., a constant fraction of slots are taken). We set the hash table size to the number of edges in the initial CSR. This guarantees that when updating the CSR, there must be a comparable number of edge insertions, and thus the cost can be amortized.
However, the belated combination of and also requires all edges in to be visible to later graph traversal operations. To do this, when inserting an edge , we use as the key and as the value. Using only the first endpoint as the key allows us to look up the neighbors of a given vertex by linear probing, and also has better cache locality. When traversing all neighbors of in the overlay graph, we first process all neighbors of in the CSR . We then further look up in the hash table for , and use linear probing to get all edges incident on . These provide the additional neighbors of w.r.t. shortcuts associated with that have not yet been incorporated into .
To identify when the hash table is overloaded, at the end of each round, we use random sampling to estimate the size of the hash table. If the hash table is filled up by a constant fraction, a merge of with is triggered. During this process, we also remove all edges incident to the vertices that have been contracted. In other words, the edge removals in line 6 of Alg. 6 are also performed lazily until the next update on . Finally, we clear the hash table.
The final edges in the CH, and are also maintained by the phase-concurrent hash tables. For both of them, only the insertion operation is needed, which can be performed concurrently using atomic operations.
Maintaining the Distance Mapping . Another important data structure used in our algorithm is , which memoizes shortest distances for certain pairs of vertices discovered in WPSes. In our implementation, we also use a phase-concurrent hash table. When a shortest distance between and is computed as , we add to the hash table, where is the key and is the value. By setting up the mapping in the LocalSearch step, later Score and Contract steps can directly look up the distance instead of performing another simulated contraction.
Similar to the shortcut set , the number of elements in the map can also change dramatically in different rounds. When most slots in the hash table are empty, clearing the map becomes expensive. To address this issue, similar to how we maintain , we clear the hash table for lazily. At the end of the round, we use sampling to estimate the load factor of the hash table. Only when a constant fraction of the hash table is full do we clear the hash table.
5. Query on Contraction Hierarchies
In this section, we describe the standard query algorithm on the CH. After constructing the CH, we re-index and sort all vertices by their level and score. In Geisberger’s sequential CH algorithm (Geisberger et al., 2008), vertices are reordered based on the order to contract them. is the order to contract . In PHAST (Delling et al., 2013b), vertices are re-grouped by levels to perform SSSP queries on CH, where vertices in lower levels always have a lower rank than those in higher levels. In the query phase, all the edges in are divided into two parts and forming two graphs: the out-edges from to higher-ranked vertices are referred to as the upward edges: , while the in-edges from higher-ranked vertices to as the downward edges: . For each edge connecting two vertices, only the endpoint of the lower rank stores this edge in .
CH boost up the efficiency of point to point query (- query) significantly. During a point-to-point query from to , a forward search from is performed on and a backward search from is performed on , meaning that only edges leading to higher rank vertices are considered. For each vertex , the distance from to and the distance from to is maintained as and and the estimated distance from to through is . Once the tentative shortest distance from to is not larger than the minimum value of the priority queue, the search result is settled to .
6. Experiments
| Graph | Road | Synthetic | -NN | ||||||||||||||
| CA | AO | AF | NA | AS | EU | TRCE | BUB | CHN7 | CHN8 | HT5 | CH2 | CH5 | GL2 | GL5 | GL10 | ||
| # vertices | 3.34M | 6.22M | 33.5M | 87.0M | 95.7M | 131M | 16.0M | 21.2M | 10.0M | 100M | 929K | 4.21M | 4.21M | 24.9M | 24.9M | 24.9M | |
| # edges | 4.20M | 8.24M | 44.8M | 113M | 123M | 169M | 48.0M | 63.6M | 10.0M | 100M | 4.64M | 8.42M | 21.0M | 49.8M | 124M | 249M | |
| Build (s) | Ours* | 1.00 | 2.25 | 7.32 | 23.1 | 23.5 | 32.9 | 25.2 | 34.7 | 1.77 | 14.9 | 0.68 | 0.87 | 115 | 2.95 | 9.78 | 41.6 |
| OSRM* | 4.60 | 26.8 | 60.0 | 307 | 286 | 521 | 218 | 292 | 6.77 | 71.2 | 2.95 | 4.38 | 28028 | 30.7 | 103 | 391 | |
| CC* | 19.0 | 32.9 | 334 | 1527 | 1548 | 2121 | 1218 | 1680 | 19.0 | 218 | 11.4 | 11.0 | 4737 | 83.3 | 453 | 2627 | |
| RK | 22.1 | 94.7 | 466 | 2466 | 2619 | 3783 | 1765 | 2509 | 38.7 | 848 | 11.2 | 15.5 | 14048 | 108 | 716 | 4175 | |
| PHAST | 15.5 | 66.2 | 287 | 1341 | 1421 | 1865 | 1470 | 2053 | 23.5 | 285 | 8.79 | 9.20 | 11361 | 46.0 | 373 | 2388 | |
| Ours-seq | 29.3 | 89.6 | 383 | 1411 | 1347 | 1823 | 1816 | 2620 | 65.4 | 751 | 18.9 | 26.1 | 7454 | 125 | 545 | 2676 | |
| OSRM-seq | 33.1 | 183 | 503 | 3246 | 2839 | 4101 | 3817 | 5273 | 34.4 | 390 | 28.4 | 18.3 | ¿30000 | 82.4 | 1062 | 12082 | |
| CC-seq | 29.2 | 119 | 553 | 2907 | 2604 | 3574 | 3658 | 5002 | 35.1 | 423 | 23.7 | 20.5 | 17268 | 133 | 812 | 6855 | |
| Query (s) | *Ours | 7.99 | 27.9 | 17.2 | 93.3 | 60.4 | 137 | 376 | 452 | 5.59 | 7.74 | 6.16 | 3.55 | 244 | 1.93 | 10.2 | 33.9 |
| OSRM* | 13.6 | 43.8 | 28.7 | 163 | 97.6 | 223 | 454 | 512 | 7.94 | 9.74 | 6.37 | 3.84 | 325 | 2.20 | 12.5 | 39.0 | |
| CC* | 22.1 | 96.4 | 48.1 | 317 | 190 | 424 | 910 | 1010 | 7.60 | 9.07 | 6.73 | 3.33 | 783 | 2.11 | 17.2 | 85.2 | |
| RK | 6.65 | 24.0 | 13.6 | 79.1 | 47.9 | 102 | 292 | 337 | 4.93 | 6.33 | 4.63 | 2.69 | 176 | 1.52 | 6.25 | 19.0 | |
| PHAST | 12.6 | 39.2 | 26.7 | 138 | 84.4 | 169 | 430 | 482 | 7.73 | 9.23 | 6.98 | 4.09 | 278 | 2.37 | 11.8 | 33.5 | |
6.1. Experimental Setup
Environment. We run our experiments on a machine machine with four Intel Xeon Gold 6252 CPUs (96 cores and 192 hyperthreads) and 1.5TB of main memory. We use numactl -i all for parallel tests to interleave the memory pages across CPUs in a round-robin fashion.
We implement our algorithms in C++ using ParlayLib (Blelloch et al., 2020a) for fork-join parallelism and parallel primitives, which is recently used in many papers on parallel algorithms (Gu et al., 2023; Shen et al., 2022; Ben-David et al., 2016; Blelloch and Dobson, 2022). ParlayLib is an algorithmic library with parallel building blocks, and also include a scheduler for support fork-join parallelism. SPoCH uses some of them such as sorting from ParlayLib, and uses the ParlayLib scheduler by default, which gives good performance in general. ParlayLib infrastructure also supports alternative schedulers (e.g., TBB (Intel Threading Building Blocks, [n. d.]) and OpenCilk (OpenCilk, 2020)). A comparison of different schedulers for SPoCH is presented in Tab. 5.
Benchmark Dataset. To evaluate the performance of CH construction algorithms, we use 16 directed graphs. This dataset includes six road graphs, four synthetic graphs and six -NN graphs.
-
•
Road graphs are the primary use application domain for CH algorithms. Here we use Central America (CA), Australia Oceania (AO), Europe (EU), North America (NA), Asia (AS), and Africa (AF) from OpenStreetMap (OpenStreetMap contributors, 2010). The edge weights are natural weights from the source data, which are up to .
-
•
Synthetic graphs with relatively low average degrees and high diameters are selected because these characteristics are ideal for testing the performance of CH algorithms. Here we use hugebubbles-0020 (BUB) and hugetrace-0020 (TRCE) from (Sanders et al., 2014), which feature 2D adaptively refined triangular meshes. We assign random weights for BUB and TRCE from 0 to . Chain graphs have (CHN7) and (CHN8) vertices with random weights from 1 to 32. Whenever a vertex is contracted in the chain, exactly one shortcut is added. Therefore, chain graphs are primarily used to test the cost related to non-scoring steps.
-
•
-NN graphs are used to evaluate the performance among different average degrees. In -NN graphs, each vertex has out-going edges pointing to its -nearest neighbors, excluding itself. As increases, the complexity of the scoring process in CH construction also increases. We use Humidity and Temperature with (HT5) (Huerta et al., 2016), Chemical with (CH2, CH5) (Fonollosa et al., 2015; Wang et al., 2021), and GeoLife with (GL2, GL5, GL10) (Wang et al., 2021; Zheng et al., 2008). Edge weights are randomly assigned from 1 to 32.
When we compare the average performance across all graphs, we always report the geometric mean values across all graphs.
Baseline Competitors. We call all existing algorithms that we compare to as baselines. We compare SPoCH to two of the most well-received sequential implementations (RoutingKit and PHAST) and two state-of-the-art parallel open-source implementations (CH-Constructor and OSRM), described as follows.
- •
- •
- •
-
•
PHAST (Group, 2024; Delling et al., 2013b): A widely used algorithm for parallel single-source shortest paths (SSSP) using CH. Its implementation includes a sequential process for constructing the CH, and a parallel process for running SSSP. In our experiment, we only compare to their construction time, which is sequential.
The basic frameworks of both parallel baselines (CC and OSRM) are based on Vetter (Vetter, 2009), which supports node scoring and contracting in parallel. Each time they score a vertex, both algorithms always perform witness searches on all of its neighbors, and they do not record any distance information from these searches. This means that if scoring two vertices involves performing WPSes on the same vertex, both implementations will perform this search twice, greatly increasing the total computation.
Since our main focus is on the performance of constructing CH, we obtain the CH generated from all baselines, and always use the same query algorithm on these CH structures to enable a fair comparison of the CH quality. We choose to use the query algorithm in RK since we observe that it has the best performance in general.
6.2. Overall Performance
Setup. We conducted experiments to compare performance in terms of construction time and query time with established baselines. Lower time consumption in both indicates better algorithm quality. Tab. 4 presents a comparison among SPoCH and all baselines. In Fig. 4, we normalize the performance of all baselines to that of SPoCH (with SPoCH’s numbers always set to 1). For construction, we also include the sequential running times of all parallel implementations in Tab. 4 to reflect the total work. For queries, for each graph, we randomly select pairs of vertices, run these queries sequentially, and report the average time and the average number of vertices processed in the queries. As mentioned, we apply the same query algorithm to all baselines.
Note that different construction algorithms may generate different CHs on the same input graph. In particular, a sequential algorithm always chooses the best node to contract in each iteration, while a parallel algorithm may choose to contract multiple vertices together. Conceptually, this means that many of the vertices are contracted earlier than they should be. Therefore, the quality of the CH may be sacrificed due to parallel construction. To measure the quality of the CHs generated by different algorithms, we use the query time as an indicator. More precisely, since all algorithms are using the same query algorithm, a lower query time indicates that the CH structure itself has higher quality.
Construction Time Evaluation. Fig. 4 and Tab. 4 show that our approach achieves significantly better construction performance than all baselines, while maintaining high CH graph quality. Even our sequential construction time is competitive to the highly-optimized sequential baselines, which is within 1.7 slower than RK and 2.9 than PHAST, and can also be faster than them by up to 2.1. This indicates that our algorithm incurs very small overhead in work to enable parallelism. When running in parallel, the construction time of our algorithm is significantly faster than all baselines. Compared to the best parallel baseline OSRM, SPoCH is 3.8–243 faster in construction, with an average of 9.9 faster. CC suffers from scalability issue in the Reset step in each round (we will discuss more in Sec. 6.3 and Sec. 6.4). As a result, it achieves the best running time with 8 threads, which is close to the performance on sequential algorithms. In general, SPoCH is 10.7–66.1 faster than CC, with an average of 31.5 faster. As shown in Fig. 7, even only considering the time for scoring and contracting, CC is still not competitive to OSRM and is much slower than SPoCH. In general, SPoCH is at least 3.8 faster than all baselines on all graphs. On average across all graphs, SPoCH is 9.9 faster than OSRM, 31.5 faster than CC, 32.4 faster than (sequential) PHAST, and 53.8 than (sequential) RK.
We then analyze the query performance. As mentioned, the quality for a CH generated by a parallel algorithm is generally expected to be lower than those generated sequentially. The results indicate that our algorithm still achieves competitive quality to the best sequential implementation. In general, RK achieves the best query performance in both time and number of vertices visited. SPoCH is competitive to RK, with query times in 1.17–1.79, and is better than all other baselines in query time, even including the sequential algorithm PHAST. On average, our query time is 1.31 slower than RK, but is 1.25 faster than PHAST, 1.35 faster than OSRM, and 2.07 faster than CC. This indicates the CH generated by our parallel algorithm has close quality to a highly-optimized sequential algorithm, and is better than other parallel versions.
Number of CH Edges, Query Time, and Query Iterations. To compare the number of CH edges, query time, and query iterations of our algorithm against the baselines (OSRM (Luxen and Vetter, 2011), CC (Bühler and Nusser, 2024), RK (Dibbelt et al., 2016), PHAST (Delling et al., 2013b)), we present a heatmap in Fig. 5. Here “query iterations” refers to the average number of vertices visited in an - query, which is machine-independent and roughly indicates the query cost. The numbers in the heatmap are normalized to those of our algorithm. For CH edges, the differences between the baselines and our algorithm are all within 20%, indicating that the space required to store the output CH is similar. For query time and iterations, RK and PHAST, being sequential and following a stricter contraction order, perform slightly better than the parallel implementations. In the worst case, our algorithm is only about twice as slow as the fastest query time. However, since queries finish in microseconds ( seconds), this twofold slowdown is negligible. Among all parallel implementations, our algorithm achieves the best average query time and iterations. In summary, our algorithm offers significantly faster construction performance while maintaining competitive output graph size and query time.
6.3. Scalability
Self-Relative Speedup. We first show the self-relative speedup of our algorithms on all the graphs in Fig. 6. We vary the number of processors from to h ( hyperthreads). Our self-relative speedup is – on road graphs, – on synthetic graphs and – on -NN graphs. This shows that our algorithm achieves high parallelism on all tested graphs.
Scalability Breakdown. To validate our claims of improving parallelism in CH preprocessing and to illustrate the limitations of existing state-of-the-art parallel implementations, we test the running times of three parallel algorithms—SPoCH, CC, and OSRM—with cores varying from to h ( hyperthreads) in Fig. 7. We select three graphs: AF, CHN7, and GL5, each from a different category. As mentioned, all parallel implementations roughly follow Vetter’s algorithm, so we split the total time into three parts: Score (running WPSes and recomputing vertex scores), Contracting (finding an independent set, contracting them and adding shortcuts), and Others.
CC does not achieve satisfactory scalability, and achieves its best performance with 8 threads. The major issue lies in a Reset step that clears arrays at the beginning of each round. This cost increases drastically with the number of threads. To clarify the breakdown between Score and Contract, we also draw a figure for CC without the Others part, referred to as . The sequential running time for the three algorithms are close. However, the running time for both CC and OSRM flattens after 16 threads. Especially for CC, even without considering the resetting time in the Others part, its time in Contract remains the same or even increases when more than 16 threads are used. This is mainly due to the use of a lock-based structure when adding shortcuts, since using more threads may result in higher contention and further degrade the performance. Our solution using parallel hash table effectively avoids this issue, and remains scalable until using all 192 (hyper)threads.
OSRM has reasonable scalability up to 16 threads, and its running time remains the same when more threads are used. One possible reason lies in the parallel granularity for computing WPS—OSRM employs parallelism on the level of all vertices in (vertices that require score recomputing). In particular, all vertices in are parallelized, but each vertex itself is processed sequentially, which includes a total of executions of Dijkstra’s algorithm. In later rounds when most vertices have been contracted, there are only a few vertices in , but their degree may have become large, causing a huge amount of work executed sequentially. Therefore, its poor performance is due to insufficient parallelism in later rounds. This issue is also reflected in Fig. 8, which we discuss in Sec. 6.4. In this case, the WPSes are close to a sequential execution, and do not benefit from having more threads. Our solution that batches all WPS sources and parallelizes them as a whole effectively avoids this issue, and can efficiently utilize more threads.
On the three tested graphs, SPoCH shows high scalability up to 96h, and always has better performance with more threads. The self-relative speedup is on AF, on CHN7, and on GL5. All techniques in SPoCH are carefully optimized for high parallelism, such as using lock-free data structures and batching WPSes to run in parallel. Therefore, with a reasonable sequential cost, the good scalability guarantees low parallel running time.
6.4. In-Depth Study for Parallel Algorithms
Recall that parallel CH construction algorithms select an independent set in each round, contract them all in parallel, and repeat until the graph becomes empty. Therefore, the performance and statistics of each round may help to better understand the process of parallel algorithms. In this section, we present an in-depth study for the performance breakdown in each round, for all parallel implementations SPoCH, CC and OSRM. We present the results on four graphs in Fig. 8, which includes the number of vertices processed and accumulated running time up to each round.
Number of Vertices Processed. In Fig. 8, we present the accumulated numbers of vertices left in the overlay graph (), the number of vertices contracted (), as well as the number of WPSes (i.e., the number of executions of Dijkstra’s algorithm). We first note that OSRM incurs much more rounds than SPoCH and CC—up to 4 more rounds than SPoCH. The larger number of rounds for OSRM is caused by how the independent set is selected. OSRM selects a vertex when it has the minimum score in its 2-hop neighborhood, while SPoCH uses 1-hop neighborhood. The goal of using 2-hop neighborhood is to avoid using complicated concurrent data structure to deal with shortcut insertion, but it results in a much slower contraction process than SPoCH. On the contrary, SPoCH uses the efficient data structure introduced in Sec. 4, which allows contracting all vertices with minimum score in their 1-hop neighborhood, leading to fewer rounds.
For all three algorithms, the size of the overlay graph decreases quickly. On CHN8, the trend of decreases linearly on the log scale. Indeed on a chain, since almost all vertices have the same edge difference, the contraction is mostly determined by the random priority. In this case, a constant fraction of vertices can be selected in expectation (Shun et al., 2015). On other graphs, the size of the overlay graph decreases rapidly in the first few rounds, and then slows down. In general, for all implementations, the overlay graph size decreases by more than 99% in the first 10–20 rounds. This implies great potential of parallelism in the construction of CH.
We then compare the number of WPSes executed for each algorithm. As mentioned, one of our efforts is to reduce the number of WPSes performed to optimize the overall performance. SPoCH incurs significantly fewer Witness Path Searches (WPSes) compared to CC and OSRM. In each round, especially in later rounds, SPoCH requires far fewer WPSes than the total number of vertices in the overlay graph. This reduction in the number of WPS demonstrates the efficiency of our approach in limiting unnecessary searches.
Performance Breakdown. We now analyze the accumulated running time across rounds. In the first of the rounds, the running time for both algorithms grows quickly. As mentioned, more than 99% of the vertices may be contracted just in the first 10–20 rounds. As the algorithm proceeds, the overlay graph becomes much smaller. On the four graphs, all algorithms reach a small overlay graph with within 30 rounds, after which very little computation is required. Indeed, our algorithm spends the majority of its time in the first several rounds. As the size of the overlay graph decreases, the running time for the last 90% of rounds is small. Indeed, all the steps become cheap after a few rounds. This indicates the effectiveness of our solution to maintain the overlay graph—when the overlay graph becomes small, the maintenance cost also shrinks proportionally. This benefit comes from our technique of maintaining the overlay graph in a lazy manner. In later rounds where only a few shortcuts are generated in each round, SPoCH delays the process of combining them into the CSR until it collects a sufficient number of shortcuts. Thus, after – rounds, there are likely only one or two global merges incurred, leading to SPoCH’s high overall performance.
In contrast, the running time for OSRM grows steadily across all rounds. This indicates that even processing a small overlay graph, OSRM may spend a noticeable amount of time. As we discussed in Sec. 6.3, this is likely due to insufficient parallelism in later rounds. For CC, as mentioned, the use of inefficient data structures to maintain shortcuts dramatically increases its running time in the Contract step, leading to overall unsatisfactory performance.
In summary, our design allows for contracting a large fraction of vertices in each round, leading to a very fast contraction process. Our new design for the LocalSearch step also reduces the number for WPSes performed. Our data structure also avoids a high cost to maintain a small overlay graph. As a result, SPoCH outperforms both parallel baselines due to new designs in both algorithm.
| Graph | Road | Synthetic | -NN | ||||||||||||||
|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|
| CA | AO | AF | NA | AS | EU | TRCE | BUB | CHN7 | CHN8 | HT5 | CH2 | CH5 | GL2 | GL5 | GL10 | ||
| Build | ParlayLib (Blelloch et al., 2020a) | 1.00 | 2.25 | 7.32 | 23.1 | 23.5 | 32.9 | 25.2 | 34.7 | 1.77 | 14.9 | 0.68 | 0.87 | 115 | 2.95 | 9.78 | 41.6 |
| OpenCilk (OpenCilk, 2020) | 0.97 | 2.21 | 7.37 | 23.1 | 23.7 | 32.5 | 24.9 | 32.4 | 1.82 | 15.4 | 0.67 | 0.86 | 111 | 2.96 | 9.71 | 41.2 | |
| TBB (Intel Threading Building Blocks, [n. d.]) | 1.03 | 2.33 | 7.66 | 23.5 | 24.0 | 32.8 | 25.1 | 34.3 | 1.93 | 15.9 | 0.50 | 0.96 | 114 | 3.30 | 10.3 | 42.3 | |
| Query | ParlayLib | 7.99 | 27.9 | 17.2 | 93.3 | 60.4 | 137 | 376 | 452 | 5.59 | 7.74 | 6.16 | 3.55 | 244 | 1.93 | 10.2 | 33.9 |
| OpenCilk | 7.91 | 27.8 | 17.1 | 93.3 | 60.7 | 141 | 379 | 454 | 5.56 | 7.78 | 6.24 | 3.53 | 239 | 1.93 | 10.4 | 32.8 | |
| TBB | 7.90 | 27.9 | 17.3 | 94.4 | 57.6 | 136 | 377 | 452 | 5.64 | 7.76 | 6.27 | 3.54 | 242 | 1.92 | 10.3 | 32.5 | |
Evaluating the Lazy Combination of Shortcut Edges. As mentioned, one of our optimizations is to combine the shortcuts lazily with the edges of the overlay graph . This is to reduce and amortize the cost of updating the CSR for . Fig. 9 shows the running time with and without this optimization. For each graph, the left bar represents the breakdown that combines in each round (without lazy combination), while the right bar represents that with the optimization. The cost is counted in the Contract step. By combining the shortcuts lazily instead of doing it every round, the cost of the Contract step is reduced from 39.9% to 19.8% of the overall time on average. Across all graphs, this optimization improves the performance by 1.1–2.7 for the total running time.
Impact of Different Schedulers. To evaluate the impact of different schedulers to the performance of SPoCH, we conduct an experiment comparing ParlayLib (Blelloch et al., 2020a), OpenCilk (OpenCilk, 2020), and TBB (Intel Threading Building Blocks, [n. d.]). Our algorithmic framework remains the same across all tests, ensuring that any observed performance differences stem primarily from the underlying scheduler rather than the core algorithm itself. Across all schedulers, the average build time is within of the fastest result, and the average query time is within . Hence, the choice of scheduler has a negligible effect on the performance of SPoCH. We present detailed running times in Tab. 5.
7. Related Work
Computing the shortest paths on a graph is one of the most well-studied problems in computer science. We refer the audience to the excellent surveys on this topic, including those by Bast et al. (Bast et al., 2016), Madkoure et al. (Madkour et al., 2017), and Sommer (Sommer, 2014). These surveys all discuss contraction hierarchies (CH) in detail. Meanwhile, we note that the main applications of the CH are on sparse networks such as roadmaps (transition networks). We acknowledge that many interesting algorithms have been proposed for relevant but different problems, such as on social networks, queries among all pairs or between a batch of sources and destinations, on dynamic graphs, and time-dependant queries. A brief list of recent work on computing shortest paths includes (Zhang et al., 2022, 2021; Qiu et al., 2022; Li et al., 2020a; Yu et al., 2020; Li et al., 2020b; Zhang and Yu, 2020; Huang et al., 2021; Gong et al., 2024).
The idea of CH was proposed by Geisberger et al. (Geisberger et al., 2008), based on simplifying highway hierarchies (Sanders and Schultes, 2006; Knopp et al., 2007) and highway node routing (Schultes and Sanders, 2007). CH has achieved notable success in practical applications, and fostered numerous later studies on relevant problems. Examples include the time-dependent versions (Batz et al., 2010, 2013), parallel and distributed versions (Kieritz et al., 2010; Vetter, 2009), on dynamic graphs (Ouyang et al., 2020), and more algorithmic optimizations (Geisberger et al., 2012; Delling et al., 2017; Hespe and Sanders, 2019). CH is also used to parallelize SSSP queries, such as in PHAST (Group, 2024; Delling et al., 2013b). Another stream of research is to derive theoretical guarantees for CH (Dibbelt et al., 2016; Blum et al., 2021; Funke and Storandt, 2015; Abraham et al., 2010; Hamann and Strasser, 2018; Columbus and Bauer, 2009). These analyses are mostly parameterized, based on some graph invariants such as tree depths, tree widths, and diameters.
There have been many studies on parallelizing CH construction and relevant techniques. Vetter’s work (Vetter, 2009) is the earliest and inspired many of the later studies (Luxen and Vetter, 2024, 2011; Chen et al., 2024; Karimi et al., 2019, 2020; Kieritz et al., 2010; Bühler and Nusser, 2024). We reviewed Vetter’s approach in Sec. 2.2. OSRM by Luxen and Vetter (Luxen and Vetter, 2024, 2011) implements Vetter’s approach, and it is considered the SOTA parallel CH construction. CH-Constructor (Bühler and Nusser, 2024) is another open-source software for parallel CH construction. We compared to OSRM and CH-Constructor since they have open-source code available. The others focus on different settings, such as distributed (Kieritz et al., 2010), GPU (Karimi et al., 2019, 2020), and on edge contraction (Chen et al., 2024). Among them, an existing GPU algorithm (Karimi et al., 2020) also prune unnecessary shortcuts. SPoCH has the same motivation for pruning, but uses a different methodology by leveraging memoization to further reduce the cost. We believe that some of our algorithmic techniques, such as LocalSearch with memoization, are applicable to other settings such as GPUs. However, we acknowledge that part of the implementation, such as the parallel data structures, will require careful redesign, which we leave as future work.
We note that there are other shortest-path algorithms that can provide different construction-query trade-offs or are designed for other graph types. Some of these algorithms include transit node routing (Bast et al., 2007), hub labeling (Abraham et al., 2011, 2012), pruned landmark labeling (Akiba et al., 2013), highway labeling (Akiba et al., 2014), and ALT (Goldberg and Harrelson, 2005). Indeed, many of these approaches (Arz et al., 2013; Bast et al., 2006, 2007; Abraham et al., 2011, 2012; Delling et al., 2013a; Bauer et al., 2010; Delling et al., 2013b) use CH as a subroutine. We believe that the faster CH construction presented in this paper can improve such trade-offs, and we leave it as future work.
8. Conclusions and Future Work
In this paper, we propose SPoCH (Scalable Parallelization of Contraction Hierarchies), a parallel algorithm for constructing contraction hierarchies. Our key insights include algorithm redesign to introduce the LocalSearch step, which allows for batching, memoization and pruning, as well as leveraging parallel data structures. In this way, SPoCH effectively reduces the total work and enhances parallelism. Across 16 graphs of varying sizes and average degrees (including road networks, synthetic graphs, and -NN graphs), SPoCH consistently outperforms four other SOTA sequential and parallel baselines by 3.83–243, while maintaining competitive query performance. On a 96-core machine, SPoCH delivers self-relative speedups of 27.9–70.8. We conduct in-depth experiments to analyze the improvements of our techniques.
An open question remains as to whether parallel CHs can handle graphs with high degrees (e.g., social networks). Some interesting future directions include applying our new CH construction algorithm to other distance queries, such as single-source shortest paths, all-pairs shortest paths, or distance oracles.
Acknowledgements.
This work is supported by NSF grants CCF-2103483, IIS-2227669, and TI-2346223, NSF CAREER Awards CCF-2238358 and CCF-2339310, the UCR Regents Faculty Development Award, and the Google Research Scholar Program. We thank the anonymous reviewers for their useful comments.References
- (1)
- Abeywickrama et al. (2016) Tenindra Abeywickrama, Muhammad Aamir Cheema, and David Taniar. 2016. k-Nearest Neighbors on Road Networks: A Journey in Experimentation and In-Memory Implementation. Proceedings of the VLDB Endowment (PVLDB) 9, 6 (2016).
- Abraham et al. (2011) Ittai Abraham, Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2011. A hub-based labeling algorithm for shortest paths in road networks. In International Symposium on Experimental Algorithms (SEA). Springer, 230–241.
- Abraham et al. (2012) Ittai Abraham, Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2012. Hierarchical hub labelings for shortest paths. In European Symposium on Algorithms (ESA). Springer, 24–35.
- Abraham et al. (2010) Ittai Abraham, Amos Fiat, Andrew V Goldberg, and Renato F Werneck. 2010. Highway dimension, shortest paths, and provably efficient algorithms. In ACM-SIAM Symposium on Discrete Algorithms (SODA). SIAM, 782–793.
- Akiba et al. (2014) Takuya Akiba, Yoichi Iwata, Ken-ichi Kawarabayashi, and Yuki Kawata. 2014. Fast shortest-path distance queries on road networks by pruned highway labeling. In Algorithm Engineering and Experiments (ALENEX). SIAM, 147–154.
- Akiba et al. (2013) Takuya Akiba, Yoichi Iwata, and Yuichi Yoshida. 2013. Fast exact shortest-path distance queries on large networks by pruned landmark labeling. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 349–360.
- Arora et al. (2001) Nimar S Arora, Robert D Blumofe, and C Greg Plaxton. 2001. Thread scheduling for multiprogrammed multiprocessors. Theory of Computing Systems (TOCS) 34, 2 (2001), 115–144.
- Arz et al. (2013) Julian Arz, Dennis Luxen, and Peter Sanders. 2013. Transit node routing reconsidered. In International symposium on experimental algorithms. Springer, 55–66.
- Bast et al. (2016) Hannah Bast, Daniel Delling, Andrew Goldberg, Matthias Mü ller Hannemann, Thomas Pajor, Peter Sanders, Dorothea Wagner, and Renato F Werneck. 2016. Route planning in transportation networks. Algorithm engineering: Selected results and surveys (2016), 19–80.
- Bast et al. (2006) Hannah Bast, Stefan Funke, and Domagoj Matijevic. 2006. Ultrafast Shortest-Path Queries via Transit Nodes. The Shortest Path Problem 74 (2006), 175–192.
- Bast et al. (2007) Holger Bast, Stefan Funke, Peter Sanders, and Dominik Schultes. 2007. Fast routing in road networks with transit nodes. Science 316, 5824 (2007), 566–566.
- Batz et al. (2010) Gernot Veit Batz, Robert Geisberger, Sabine Neubauer, and Peter Sanders. 2010. Time-dependent contraction hierarchies and approximation. In International Symposium on Experimental Algorithms (SEA). Springer, 166–177.
- Batz et al. (2013) G Veit Batz, Robert Geisberger, Peter Sanders, and Christian Vetter. 2013. Minimum time-dependent travel times with contraction hierarchies. Journal of Experimental Algorithmics (JEA) 18 (2013), 1–1.
- Bauer et al. (2010) Reinhard Bauer, Daniel Delling, Peter Sanders, Dennis Schieferdecker, Dominik Schultes, and Dorothea Wagner. 2010. Combining hierarchical and goal-directed speed-up techniques for dijkstra’s algorithm. Journal of Experimental Algorithmics (JEA) 15 (2010), 2–1.
- Ben-David et al. (2016) Naama Ben-David, Guy E. Blelloch, Jeremy T. Fineman, Phillip B. Gibbons, Yan Gu, Charles McGuffey, and Julian Shun. 2016. Parallel Algorithms for Asymmetric Read-Write Costs. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Blelloch et al. (2020a) Guy E. Blelloch, Daniel Anderson, and Laxman Dhulipala. 2020a. ParlayLib — a toolkit for parallel algorithms on shared-memory multicore machines. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 507–509.
- Blelloch and Dobson (2022) Guy E Blelloch and Magdalen Dobson. 2022. Parallel Nearest Neighbors in Low Dimensions with Batch Updates. In Algorithm Engineering and Experiments (ALENEX). SIAM, 195–208.
- Blelloch et al. (2020b) Guy E. Blelloch, Jeremy T. Fineman, Yan Gu, and Yihan Sun. 2020b. Optimal parallel algorithms in the binary-forking model. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 89–102.
- Blelloch et al. (2010) Guy E. Blelloch, Phillip B. Gibbons, and Harsha Vardhan Simhadri. 2010. Low depth cache-oblivious algorithms. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Blum et al. (2021) Johannes Blum, Stefan Funke, and Sabine Storandt. 2021. Sublinear search spaces for shortest path planning in grid and road networks. Journal of Combinatorial Optimization 42, 2 (2021), 231–257.
- Blumofe and Leiserson (1999) Robert D. Blumofe and Charles E. Leiserson. 1999. Scheduling multithreaded computations by work stealing. J. ACM 46, 5 (1999), 720–748.
- Bühler and Nusser (2024) Stefan Bühler and André Nusser. 2024. Contraction Hierarchies Constructor. https://doi.org/10.5281/zenodo.14008202
- Chen et al. (2024) Zi Chen, Xinyu Ji, Long Yuan, Xuemin Lin, Wenjie Zhang, and Shan Huang. 2024. Parallel Contraction Hierarchies Construction on Road Networks. IEEE Transactions on Knowledge and Data Engineering (2024).
- Columbus and Bauer (2009) Tobias Columbus and Reinhard Bauer. 2009. On the complexity of contraction hierarchies. Student Thesis, Karlsruhe Inst. Technol., Karlsruhe, Germany (2009).
- contributors (2024) RoutingKit contributors. 2024. RoutingKit. https://github.com/RoutingKit/RoutingKit.
- Delling et al. (2013b) Daniel Delling, Andrew V Goldberg, Andreas Nowatzyk, and Renato F Werneck. 2013b. PHAST: Hardware-accelerated shortest path trees. J. Parallel and Distrib. Comput. 73, 7 (2013), 940–952.
- Delling et al. (2017) Daniel Delling, Andrew V Goldberg, Thomas Pajor, and Renato F Werneck. 2017. Customizable route planning in road networks. Transportation Science 51, 2 (2017), 566–591.
- Delling et al. (2013a) Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2013a. Hub label compression. In International symposium on experimental algorithms. Springer, 18–29.
- Dhulipala et al. (2021) Laxman Dhulipala, Guy E. Blelloch, and Julian Shun. 2021. Theoretically efficient parallel graph algorithms can be fast and scalable. ACM Transactions on Parallel Computing (TOPC) 8, 1 (2021), 1–70.
- Dhulipala et al. (2024) Laxman Dhulipala, Xiaojun Dong, Kishen Gowda, and Yan Gu. 2024. Optimal Parallel Algorithms for Dendrogram Computation and Single-Linkage Clustering. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Dibbelt et al. (2016) Julian Dibbelt, Ben Strasser, and Dorothea Wagner. 2016. Customizable contraction hierarchies. J. Experimental Algorithmics 21 (2016), 1–49.
- Dijkstra (1959) Edsger W. Dijkstra. 1959. A note on two problems in connexion with graphs. Numerische mathematik 1, 1 (1959).
- Dong et al. (2024a) Xiaojun Dong, Laxman Dhulipala, Yan Gu, and Yihan Sun. 2024a. Parallel Integer Sort: Theory and Practice. In ACM Symposium on Principles and Practice of Parallel Programming (PPOPP).
- Dong et al. (2024b) Xiaojun Dong, Yan Gu, Yihan Sun, and Letong Wang. 2024b. Brief Announcement: PASGAL: Parallel And Scalable Graph Algorithm Library. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Dong et al. (2021) Xiaojun Dong, Yan Gu, Yihan Sun, and Yunming Zhang. 2021. Efficient Stepping Algorithms and Implementations for Parallel Shortest Paths. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 184–197.
- Dong et al. (2023) Xiaojun Dong, Letong Wang, Yan Gu, and Yihan Sun. 2023. Provably Fast and Space-Efficient Parallel Biconnectivity. In ACM Symposium on Principles and Practice of Parallel Programming (PPOPP). 52–65.
- Fonollosa et al. (2015) Jordi Fonollosa, Sadique Sheik, Ramón Huerta, and Santiago Marco. 2015. Reservoir computing compensates slow response of chemosensor arrays exposed to fast varying gas concentrations in continuous monitoring. Sensors and Actuators B: Chemical 215 (2015), 618–629.
- Funke et al. (2015) Stefan Funke, André Nusser, and Sabine Storandt. 2015. Placement of loading stations for electric vehicles: No detours necessary! Journal of Artificial Intelligence Research 53 (2015), 633–658.
- Funke et al. (2019) Stefan Funke, Tobias Rupp, André Nusser, and Sabine Storandt. 2019. PATHFINDER: storage and indexing of massive trajectory sets. In Proceedings of the 16th International Symposium on Spatial and Temporal Databases. 90–99.
- Funke and Storandt (2015) Stefan Funke and Sabine Storandt. 2015. Provable efficiency of contraction hierarchies with randomized preprocessing. In International Symposium on Algorithms and Computation. Springer, 479–490.
- Geisberger et al. (2008) Robert Geisberger, Peter Sanders, Dominik Schultes, and Daniel Delling. 2008. Contraction hierarchies: Faster and simpler hierarchical routing in road networks. In International Workshop on Experimental Algorithms (WEA). Springer, 319–333.
- Geisberger et al. (2012) Robert Geisberger, Peter Sanders, Dominik Schultes, and Christian Vetter. 2012. Exact routing in large road networks using contraction hierarchies. Transportation Science 46, 3 (2012), 388–404.
- Goldberg and Harrelson (2005) Andrew V Goldberg and Chris Harrelson. 2005. Computing the shortest path: A search meets graph theory. In ACM-SIAM Symposium on Discrete Algorithms (SODA), Vol. 5. 156–165.
- Gong et al. (2024) Zengyang Gong, Yuxiang Zeng, and Lei Chen. 2024. Querying shortest path on large time-dependent road networks with shortcuts. In IEEE International Conference on Data Engineering (ICDE). IEEE, 4532–4544.
- Group (2024) KIT ITI Algorithmics Group. 2024. PHAST. https://github.com/kit-algo/ULTRA-PHAST.
- Gu et al. (2023) Yan Gu, Ziyang Men, Zheqi Shen, Yihan Sun, and Zijin Wan. 2023. Parallel Longest Increasing Subsequence and van Emde Boas Trees. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Hamann and Strasser (2018) Michael Hamann and Ben Strasser. 2018. Graph bisection with pareto optimization. J. Experimental Algorithmics 23 (2018), 1–34.
- Hespe and Sanders (2019) Demian Hespe and Peter Sanders. 2019. More Hierarchy in Route Planning Using Edge Hierarchies. In Symposium on Algorithmic Approaches for Transportation Modelling, Optimization, and Systems (ATMOS).
- Huang et al. (2021) Shuai Huang, Yong Wang, Tianyu Zhao, and Guoliang Li. 2021. A learning-based method for computing shortest path distances on road networks. In IEEE International Conference on Data Engineering (ICDE). IEEE, 360–371.
- Huerta et al. (2016) Ramon Huerta, Thiago Mosqueiro, Jordi Fonollosa, Nikolai F Rulkov, and Irene Rodriguez-Lujan. 2016. Online decorrelation of humidity and temperature in chemical sensors for continuous monitoring. Chemometrics and Intelligent Laboratory Systems 157 (2016), 169–176.
- Intel Threading Building Blocks ([n. d.]) Intel Threading Building Blocks [n. d.]. Intel Threading Building Blocks (TBB). https://www.threadingbuildingblocks.org.
- Karimi et al. (2019) Roozbeh Karimi, David M Koppelman, and Chris J Michael. 2019. GPU road network graph contraction and SSSP query. In Proceedings of the ACM International Conference on Supercomputing. 250–260.
- Karimi et al. (2020) Roozbeh Karimi, David M Koppelman, and Chris J Michael. 2020. Fast GPU Graph Contraction by Combining Efficient Shallow Searches and Post-Culling. In IEEE Conference on High Performance Extreme Computing (HPEC). IEEE, 1–7.
- Kieritz et al. (2010) Tim Kieritz, Dennis Luxen, Peter Sanders, and Christian Vetter. 2010. Distributed time-dependent contraction hierarchies. In International Symposium on Experimental Algorithms (SEA). Springer, 83–93.
- Knopp et al. (2007) Sebastian Knopp, Peter Sanders, Dominik Schultes, Frank Schulz, and Dorothea Wagner. 2007. Computing many-to-many shortest paths using highway hierarchies. In Algorithm Engineering and Experiments (ALENEX). SIAM, 36–45.
- Li et al. (2020a) Lingxiao Li, Muhammad Aamir Cheema, Mohammed Eunus Ali, Hua Lu, and David Taniar. 2020a. Continuously monitoring alternative shortest paths on road networks. In Proceedings of the VLDB Endowment (PVLDB). Association for Computing Machinery, 2243–2255.
- Li et al. (2020b) Lei Li, Mengxuan Zhang, Wen Hua, and Xiaofang Zhou. 2020b. Fast query decomposition for batch shortest path processing in road networks. In IEEE International Conference on Data Engineering (ICDE). IEEE, 1189–1200.
- Liu et al. (2025) Youzhe Liu, Xiaojun Dong, Yan Gu, and Yihan Sun. 2025. Parallel -Core Decomposition: Theory and Practice. ACM SIGMOD International Conference on Management of Data (SIGMOD) 3, 3 (2025).
- Luxen and Vetter (2011) Dennis Luxen and Christian Vetter. 2011. Real-time routing with OpenStreetMap data. In SIGSPATIAL international conference on advances in geographic information systems. 513–516.
- Luxen and Vetter (2024) Dennis Luxen and Christian Vetter. 2024. Open Source Routing Machine. https://github.com/Project-OSRM/osrm-backend.
- Madkour et al. (2017) Amgad Madkour, Walid G Aref, Faizan Ur Rehman, Mohamed Abdur Rahman, and Saleh Basalamah. 2017. A survey of shortest-path algorithms. arXiv preprint:1705.02044 (2017).
- Meyer and Sanders (2003) Ulrich Meyer and Peter Sanders. 2003. -stepping: a parallelizable shortest path algorithm. Journal of Algorithms 49, 1 (2003), 114–152.
- OpenCilk (2020) OpenCilk 2020. OpenCilk. https://www.opencilk.org/.
- OpenStreetMap contributors (2010) OpenStreetMap contributors. 2010. OpenStreetMap. https://www.openstreetmap.org/.
- Ouyang et al. (2020) Dian Ouyang, Long Yuan, Lu Qin, Lijun Chang, Ying Zhang, and Xuemin Lin. 2020. Efficient shortest path index maintenance on dynamic road networks with theoretical guarantees. Proceedings of the VLDB Endowment (PVLDB) 13, 5 (2020), 602–615.
- Qiu et al. (2022) Yu-Xuan Qiu, Dong Wen, Lu Qin, Wentao Li, Rong-Hua Li, Zhang Ying, et al. 2022. Efficient shortest path counting on large road networks. Proceedings of the VLDB Endowment (PVLDB) (2022).
- Sanchez et al. (2022) Naroa Coretti Sanchez, Iñigo Martinez, Luis Alonso Pastor, and Kent Larson. 2022. On the simulation of shared autonomous micro-mobility. Communications in Transportation Research 2 (2022), 100065.
- Sanders and Schultes (2006) Peter Sanders and Dominik Schultes. 2006. Engineering highway hierarchies. In European Symposium on Algorithms. Springer, 804–816.
- Sanders et al. (2014) Peter Sanders, Christian Schulz, and Dorothea Wagner. 2014. Benchmarking for graph clustering and partitioning. Encyclopedia of social network analysis and mining Springer (2014).
- Schultes and Sanders (2007) Dominik Schultes and Peter Sanders. 2007. Dynamic highway-node routing. In Workshop on Experimental and Efficient Algorithms (WEA). Springer, 66–79.
- Shen et al. (2022) Zheqi Shen, Zijin Wan, Yan Gu, and Yihan Sun. 2022. Many Sequential Iterative Algorithms Can Be Parallel and (Nearly) Work-efficient. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
- Shun and Blelloch (2014) Julian Shun and Guy E Blelloch. 2014. Phase-concurrent hash tables for determinism. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 96–107.
- Shun et al. (2015) Julian Shun, Yan Gu, Guy E. Blelloch, Jeremy T. Fineman, and Phillip B Gibbons. 2015. Sequential random permutation, list contraction and tree contraction are highly parallel. In ACM-SIAM Symposium on Discrete Algorithms (SODA). 431–448.
- Sommer (2014) Christian Sommer. 2014. Shortest-path queries in static networks. ACM Computing Surveys (CSUR) 46, 4 (2014), 1–31.
- Tinney and Walker (1967) William F Tinney and John W Walker. 1967. Direct solutions of sparse network equations by optimally ordered triangular factorization. Proc. IEEE 55, 11 (1967), 1801–1809.
- Vetter (2009) Christian Vetter. 2009. Parallel time-dependent contraction hierarchies. Student Research Project (2009), 134.
- Wan et al. (2024) Zijin Wan, Xiaojun Dong, Letong Wang, Enzuo Zhu, Yan Gu, and Yihan Sun. 2024. Implementations for Parallel Contraction Hierarchies. https://github.com/ucrparlay/Parallel-Contraction-Hierarchy.
- Wang et al. (2023) Letong Wang, Xiaojun Dong, Yan Gu, and Yihan Sun. 2023. Parallel Strong Connectivity Based on Faster Reachability. ACM SIGMOD International Conference on Management of Data (SIGMOD) 1, 2 (2023), 1–29.
- Wang et al. (2021) Yiqiu Wang, Shangdi Yu, Laxman Dhulipala, Yan Gu, and Julian Shun. 2021. GeoGraph: A Framework for Graph Processing on Geometric Data. ACM SIGOPS Operating Systems Review 55, 1 (2021), 38–46.
- Wheatman et al. (2024) Brian Wheatman, Xiaojun Dong, Zheqi Shen, Laxman Dhulipala, Jakub Ł\kacki, Prashant Pandey, and Helen Xu. 2024. BYO: A Unified Framework for Benchmarking Large-Scale Graph Containers. In Proceedings of the VLDB Endowment (PVLDB).
- Yu et al. (2020) Ziqiang Yu, Xiaohui Yu, Nick Koudas, Yang Liu, Yifan Li, Yueting Chen, and Dingyu Yang. 2020. Distributed processing of k shortest path queries over dynamic road networks. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 665–679.
- Zhang et al. (2022) Junhua Zhang, Wentao Li, Long Yuan, Lu Qin, Ying Zhang, and Lijun Chang. 2022. Shortest-path queries on complex networks: experiments, analyses, and improvement. Proceedings of the VLDB Endowment (PVLDB) 15, 11 (2022), 2640–2652.
- Zhang et al. (2021) U Zhang, Long Yuan, Wentao Li, Lu Qin, and Ying Zhang. 2021. Efficient label-constrained shortest path queries on road networks: A tree decomposition approach. Proceedings of the VLDB Endowment (PVLDB) (2021).
- Zhang and Yu (2020) Yikai Zhang and Jeffrey Xu Yu. 2020. Hub labeling for shortest path counting. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 1813–1828.
- Zheng et al. (2008) Yu Zheng, Like Liu, Longhao Wang, and Xing Xie. 2008. Learning transportation mode from raw gps data for geographic applications on the web. In International World Wide Web Conference (WWW). 247–256.